class Error (Exception): pass


flags_in_order = [
    'assembled',
    'assembled_into_one_contig',
    'region_assembled_twice',
    'complete_gene',
    'unique_contig',
    'scaffold_graph_bad',
    'assembly_fail',
    'variants_suggest_collapsed_repeat',
    'hit_both_strands',
    'has_variant',
    'ref_seq_choose_fail',
]


flag_bits = {flags_in_order[i]: pow(2, i) for i in range(len(flags_in_order))}


class Flag:
    def __init__(self, n=0):
        self.flags = {x: False for x in flags_in_order}
        self.set_flag(n)


    def set_flag(self, n):
        for f in self.flags:
            if flag_bits[f] & n != 0:
                self.flags[f] = True


    def add(self, f):
        self.flags[f] = True


    def to_number(self):
        n = 0
        for f in self.flags:
            if self.flags[f]:
                n += flag_bits[f]
        return n

    def __eq__(self, other):
       return type(other) is type(self) and self.__dict__ == other.__dict__


    def __str__(self):
        return str(self.to_number())


    def to_long_string(self):
        lines = []
        for f in flags_in_order:
            x_or_not = 'X' if self.flags[f] else ' '
            lines.append('[' + x_or_not + '] ' + f)
        return '\n'.join(lines)


    def has(self, s):
        return self.flags[s]




import pysam
import pyfastaq
import os
from ariba import scaffold_graph

class Error (Exception): pass

class Parser:
    def __init__(self, bam, ref_seqs):
        '''Construct a Parser.
        bam: name of BAM file
        ref_seqs: dictionary of sequence name => Fasta object'''
        self.bam = os.path.abspath(bam)
        self.soft_clipped = {}
        self.ref_lengths = {seq: len(ref_seqs[seq]) for seq in ref_seqs}
        self.scaff_graph = scaffold_graph.Graph(self.ref_lengths)
        self.unmapped_mates = {}
        self.sam_reader = pysam.Samfile(self.bam, "rb")


    def _sam_to_soft_clipped(self, sam):
        '''Returns tuple of whether or not the left and right end of the mapped read in the sam record is soft-clipped'''
        if sam.is_unmapped:
            raise Error('Cannot get soft clip info from an unmapped read')

        if sam.cigar is None or len(sam.cigar) == 0:
            return False, False
        return (sam.cigar[0][0] == 4, sam.cigar[-1][0] == 4)


    def _update_soft_clipped_from_sam(self, sam):
        ref_name = self.sam_reader.getrname(sam.tid)
        left_clip, right_clip = self._sam_to_soft_clipped(sam)

        if left_clip or right_clip:
            if ref_name not in self.soft_clipped:
                self.soft_clipped[ref_name] = {}

            if left_clip:
                p = sam.pos
                if p not in self.soft_clipped[ref_name]:
                    self.soft_clipped[ref_name][p] = [0, 0]
                self.soft_clipped[ref_name][p][0] += 1

            if right_clip:
                p = sam.aend
                if p not in self.soft_clipped[ref_name]:
                    self.soft_clipped[ref_name][p] = [0, 0]
                self.soft_clipped[ref_name][p][1] += 1


    def _update_unmapped_mates_from_sam(self, sam):
        if (not sam.is_unmapped) and sam.mate_is_unmapped:
            ref_name = self.sam_reader.getrname(sam.tid)
            if ref_name not in self.unmapped_mates:
                self.unmapped_mates[ref_name] = {}
            pos = sam.reference_start
            self.unmapped_mates[ref_name][pos] = self.unmapped_mates[ref_name].get(pos, 0) + 1


    def _write_soft_clipped_to_file(self, filename):
        f = pyfastaq.utils.open_file_write(filename)
        for seq in sorted(self.soft_clipped):
            for position in sorted(self.soft_clipped[seq]):
                print(seq, position + 1, self.soft_clipped[seq][position][0], self.soft_clipped[seq][position][1], sep='\t', file=f)
        pyfastaq.utils.close(f)


    def _write_unmapped_mates_to_file(self, filename):
        f = pyfastaq.utils.open_file_write(filename)
        for seq in sorted(self.unmapped_mates):
            for position in sorted(self.unmapped_mates[seq]):
                print(seq, position + 1, self.unmapped_mates[seq][position], sep='\t', file=f)
        pyfastaq.utils.close(f)


    def parse(self):
        for sam in self.sam_reader.fetch(until_eof=True):
            if sam.is_unmapped:
                continue

            self._update_soft_clipped_from_sam(sam)
            self.scaff_graph.update_from_sam(sam, self.sam_reader)
            self._update_unmapped_mates_from_sam(sam)


    def scaff_graph_is_consistent(self, min_coverage, max_insert):
        return self.scaff_graph.is_consistent(min_coverage, max_insert)


    def write_files(self, prefix):
        self._write_soft_clipped_to_file(prefix + '.soft_clipped')
        self._write_unmapped_mates_to_file(prefix + '.unmapped_mates')
        self.scaff_graph.write_all_links_to_file(prefix + '.scaff')




import copy
import sys
import pymummer

class Error (Exception): pass

columns = [
    'ref_name',              # 0  name of reference sequence
    'gene',                  # 1  is a gene 0|1
    'var_only',              # 2  is variant only 0|1
    'flag',                  # 3  cluster flag
    'reads',                 # 4  number of reads in this cluster
    'cluster',               # 5  name of cluster
    'ref_len',               # 6  length of reference sequence
    'ref_base_assembled',    # 7  number of reference nucleotides assembled by this contig
    'pc_ident',              # 8  %identity between ref sequence and contig
    'ctg',                   # 9  name of contig matching reference
    'ctg_len',               # 10  length of contig matching reference
    'ctg_cov',               # 11 mean mapped read depth of this contig
    'known_var',             # 12 is this a known SNP from reference metadata? 1|0
    'var_type',              # 13 The type of variant. Currently only SNP supported
    'var_seq_type',          # 14 if known_var=1, n|p for nucleotide or protein
    'known_var_change',      # 15 if known_var=1, the wild/variant change, eg I42L
    'has_known_var',         # 16 if known_var=1, 1|0 for whether or not the assembly has the variant
    'ref_ctg_change',        # 17 amino acid or nucleotide change between reference and contig, eg I42L
    'ref_ctg_effect',        # 18 effect of change between reference and contig, eg SYS, NONSYN (amino acid changes only)
    'ref_start',             # 19 start position of variant in contig
    'ref_end',               # 20 end position of variant in contig
    'ref_nt',                # 21 nucleotide(s) in contig at variant position
    'ctg_start',             # 22 start position of variant in contig
    'ctg_end',               # 23 end position of variant in contig
    'ctg_nt',                # 24 nucleotide(s) in contig at variant position
    'smtls_total_depth',     # 25 total read depth at variant start position in contig, reported by mpileup
    'smtls_alt_nt',          # 26 alt nucleotides on contig, reported by mpileup
    'smtls_alt_depth',       # 27 alt depth on contig, reported by mpileup
    'var_description',       # 28 description of variant from reference metdata
    'free_text',             # 29 other free text about reference sequence, from reference metadata
]


var_columns = [
    'known_var',
    'var_type',
    'var_seq_type',
    'known_var_change',
    'has_known_var',
    'ref_ctg_change',
    'ref_ctg_effect',
    'ref_start',
    'ref_end',
    'ref_nt',
    'ctg_start',
    'ctg_end',
    'ctg_nt',
    'smtls_total_depth',
    'smtls_alt_nt',
    'smtls_alt_depth',
    'var_description',
]

int_columns = [
    'reads',
    'ref_len',
    'ref_base_assembled',
    'ctg_len',
    'ref_start',
    'ref_end',
    'ctg_start',
    'ctg_end',
]


float_columns = [
    'ctg_cov',
    'pc_ident',
]


def header_line():
    return '\t'.join(columns)


def _samtools_depths_at_known_snps_all_wild(sequence_meta, contig_name, cluster, variant_list):
    '''Input is a known variants, as sequence_metadata object. The
       assumption is that both the reference and the assembly have the
       variant type, not wild type. The list variant_list should be a list
       of pymummer.variant.Variant objects, only contaning variants to the
       relevant query contig'''
    ref_nuc_range = sequence_meta.variant.nucleotide_range()

    if ref_nuc_range is None:
        return None

    ctg_nts = []
    ref_nts = []
    smtls_total_depths = []
    smtls_alt_nts = []
    smtls_alt_depths = []
    contig_positions = []

    for ref_position in range(ref_nuc_range[0], ref_nuc_range[1]+1, 1):
        nucmer_match = cluster.assembly_compare.nucmer_hit_containing_reference_position(cluster.assembly_compare.nucmer_hits, cluster.ref_sequence.id, ref_position)

        if nucmer_match is not None:
            # work out contig position. Needs indels variants to correct the position
            ref_nts.append(cluster.ref_sequence[ref_position])
            contig_position, in_indel = nucmer_match.qry_coords_from_ref_coord(ref_position, variant_list)
            contig_positions.append(contig_position)
            ref, alt, total_depth, alt_depths = cluster.samtools_vars.get_depths_at_position(contig_name, contig_position)
            ctg_nts.append(ref)
            smtls_alt_nts.append(alt)
            smtls_total_depths.append(total_depth)
            smtls_alt_depths.append(alt_depths)

    ctg_nts = ';'.join(ctg_nts) if len(ctg_nts) else '.'
    ref_nts = ';'.join(ref_nts) if len(ref_nts) else '.'
    smtls_alt_nts = ';'.join(smtls_alt_nts) if len(smtls_alt_nts) else '.'
    smtls_total_depths = ';'.join([str(x)for x in smtls_total_depths]) if len(smtls_total_depths) else '.'
    smtls_alt_depths = ';'.join([str(x)for x in smtls_alt_depths]) if len(smtls_alt_depths) else '.'
    ctg_start = str(min(contig_positions) + 1) if contig_positions is not None else '.'
    ctg_end = str(max(contig_positions) + 1) if contig_positions is not None else '.'

    return [str(x) for x in [
        ref_nuc_range[0] + 1,
        ref_nuc_range[1] + 1,
        ref_nts,
        ctg_start,
        ctg_end,
        ctg_nts,
        smtls_total_depths,
        smtls_alt_nts,
        smtls_alt_depths
    ]]


def _report_lines_for_one_contig(cluster, contig_name, ref_cov_per_contig, pymummer_variants):
    lines = []
    contig_length = len(cluster.assembly.sequences[contig_name])
    assert contig_length != 0

    if contig_name in ref_cov_per_contig:
        if contig_name == cluster.assembly_compare.scaff_name_matching_ref and cluster.assembly_compare.gene_matching_ref_type == 'GENE_FOUND':
            ref_cov = len(cluster.ref_sequence)
        else:
            ref_cov = ref_cov_per_contig[contig_name]
    else:
        ref_cov = 0

    common_first_columns = [
        cluster.ref_sequence.id,
        cluster.is_gene,
        cluster.is_variant_only,
        str(cluster.status_flag),
        str(cluster.total_reads),
        cluster.name,
        str(len(cluster.ref_sequence)),
        str(ref_cov),
        str(cluster.assembly_compare.percent_identities[contig_name]) if contig_name in cluster.assembly_compare.percent_identities else '0',
        contig_name,
        str(contig_length),  # 9 length of scaffold matching reference
    ]

    # it's possible that there is no read depth on an assembled contig
    if contig_name in cluster.total_contig_depths:
        common_first_columns.append(str(round(cluster.total_contig_depths[contig_name] / contig_length, 1)))
    else:
        common_first_columns.append('0')

    if cluster.ref_sequence.id in cluster.refdata.metadata and  len(cluster.refdata.metadata[cluster.ref_sequence.id]['.']) > 0:
        free_text_column = ';'.join([x.free_text for x in cluster.refdata.metadata[cluster.ref_sequence.id]['.']])
    else:
        free_text_column = ';'.join(['.'])

    remaining_samtools_variants = copy.copy(cluster.variants_from_samtools)
    if cluster.assembled_ok and contig_name in cluster.assembly_variants and len(cluster.assembly_variants[contig_name]) > 0:
        for (position, var_seq_type, ref_ctg_change, var_effect, contributing_vars, matching_vars_set, metainfo_set) in cluster.assembly_variants[contig_name]:
            if len(matching_vars_set) > 0:
                is_known_var = '1'
                known_var_change = 'unknown'
                var_type = 'SNP'
                has_known_var = '1'
                matching_vars_column = ';;;'.join([x.to_string(separator=':') for x in matching_vars_set])
            else:
                is_known_var = '0'
                known_var_change = '.'
                has_known_var = '0'
                var_type = '.'
                matching_vars_column = '.'

            variant_columns = ['.' if x is None else str(x) for x in [is_known_var, var_type, var_seq_type, known_var_change, has_known_var, ref_ctg_change, var_effect]]

            if contributing_vars is None:
                samtools_columns = [['.'] * 9]
            else:
                contributing_vars.sort(key = lambda x: x.qry_start)

                smtls_total_depth = []
                smtls_alt_nt = []
                smtls_alt_depth = []

                for var in contributing_vars:
                    if contig_name in remaining_samtools_variants:
                        remaining_samtools_variants[contig_name].discard(var.qry_start)

                    depths_tuple = cluster.samtools_vars.get_depths_at_position(contig_name, var.qry_start)

                    if depths_tuple is not None:
                        smtls_alt_nt.append(depths_tuple[1])
                        smtls_total_depth.append(str(depths_tuple[2]))
                        smtls_alt_depth.append(str(depths_tuple[3]))

                smtls_total_depth = ';'.join(smtls_total_depth) if len(smtls_total_depth) else '.'
                smtls_alt_nt = ';'.join(smtls_alt_nt) if len(smtls_alt_nt) else '.'
                smtls_alt_depth = ';'.join(smtls_alt_depth) if len(smtls_alt_depth) else '.'
                samtools_columns = [
                        str(contributing_vars[0].ref_start + 1), #ref_start
                        str(contributing_vars[0].ref_end + 1), # ref_end
                        ';'.join([x.ref_base for x in contributing_vars]), # ref_nt
                        str(contributing_vars[0].qry_start + 1),  # ctg_start
                        str(contributing_vars[0].qry_end + 1),  #ctg_end
                        ';'.join([x.qry_base for x in contributing_vars]), #ctg_nt
                        smtls_total_depth,
                        smtls_alt_nt,
                        smtls_alt_depth,
                ]


            if len(matching_vars_set) > 0:
                for matching_var in matching_vars_set:
                    if contributing_vars is None:
                        samtools_columns = _samtools_depths_at_known_snps_all_wild(matching_var, contig_name, cluster, pymummer_variants)
                    variant_columns[3] = str(matching_var.variant)

                    if matching_var.has_variant(cluster.ref_sequence) == (ref_ctg_change is not None):
                        variant_columns[4] = '0'
                    else:
                        variant_columns[4] = '1'

                    if samtools_columns is None:
                        samtools_columns = [['.'] * 9]
                    else:
                        for ctg_pos in range(int(samtools_columns[3]) - 1, int(samtools_columns[4]), 1):
                            if contig_name in remaining_samtools_variants:
                                remaining_samtools_variants[contig_name].discard(ctg_pos)

                    lines.append('\t'.join(common_first_columns + variant_columns + samtools_columns + [matching_vars_column] + [free_text_column]))
            else:
                lines.append('\t'.join(
                    common_first_columns + variant_columns + \
                    samtools_columns + \
                    [matching_vars_column] + [free_text_column]
                ))

    for contig_name in remaining_samtools_variants:
        for var_position in remaining_samtools_variants[contig_name]:
            depths_tuple = cluster.samtools_vars.get_depths_at_position(contig_name, var_position)
            if depths_tuple is not None:
                new_cols = [
                    '0',  # known_var column
                    'HET', # var_type
                    '.', '.', '.', '.', '.', '.', '.', '.', # var_seq_type ... ref_nt
                    str(var_position + 1), str(var_position + 1), # ctg_start, ctg_end
                    depths_tuple[0], # ctg_nt
                    str(depths_tuple[2]), # smtls_total_depth
                    depths_tuple[1], # smtls_alt_nt
                    str(depths_tuple[3]), # smtls_alt_depth
                    '.',
                    free_text_column,
                ]
                lines.append('\t'.join(common_first_columns + new_cols))

    if len(lines) == 0:
        lines.append('\t'.join(common_first_columns + ['.'] * (len(columns) - len(common_first_columns) - 1) + [free_text_column]))

    return lines


def report_lines(cluster):
    if cluster.status_flag.has('ref_seq_choose_fail'):
        fields = ['.', '.', '.', str(cluster.status_flag), str(cluster.total_reads), cluster.name] + ['.'] * (len(columns) - 6)
        assert len(fields) == len(columns)
        return ['\t'.join(fields)]
    elif cluster.status_flag.has('assembly_fail'):
        fields = ['.', '.', '.', str(cluster.status_flag), str(cluster.total_reads), cluster.name] + ['.'] * (len(columns) - 6)
        assert len(fields) == len(columns)
        return ['\t'.join(fields)]

    ref_cov_per_contig = cluster.assembly_compare.ref_cov_per_contig(cluster.assembly_compare.nucmer_hits)
    lines = []
    pymummer_variants = pymummer.snp_file.get_all_variants(cluster.assembly_compare.nucmer_snps_file)

    for contig_name in sorted(cluster.assembly.sequences):
        contig_pymummer_variants = [x for x in pymummer_variants if x.qry_name == contig_name]
        lines.extend(_report_lines_for_one_contig(cluster, contig_name, ref_cov_per_contig, contig_pymummer_variants))

    lines_ok = True

    for line in lines:
        if len(line.split('\t')) != len(columns):
            cols = line.split('\t')
            print('Error making report - wrong number of columns. Expected', len(columns), 'but got', len(cols), file=sys.stderr)
            for i in range(len(cols)):
                print(i, cols[i], sep='\t', file=sys.stderr)
            lines_ok = False

    if not lines_ok:
        raise Error('Error making report. Cannot continue')

    return lines if len(lines) > 0 else None




import pyfastaq
import re

class Error (Exception): pass


allowed_variant_types = {'n', 'p'}

class Variant:
    def __init__(self, variant_type, variant_string, identifier):
        if variant_type not in allowed_variant_types:
            raise Error('Error! Variant type "' + variant_type + '" not recognised.\n' + \
                        'Must be one of:' + ', '.join(allowed_variant_types))

        self.variant_type = variant_type
        self.identifier = None if identifier == '.' else identifier


        m = re.match('^([A-Z])([0-9]+)([A-Z])$', variant_string.upper())
        if m is None:
            raise Error('Unexpected format of variant string: ', variant_string)

        try:
            self.wild_value, self.position, self.variant_value = m.group(1, 2, 3)
        except:
            raise Error('Error getting amino acids and position of variant from', variant_string)

        self.position = int(self.position) - 1


    def __eq__(self, other):
       return type(other) is type(self) and self.__dict__ == other.__dict__


    def __lt__(self, other):
        return self.position < other.position or \
            (self.position == other.position and self.variant_type < other.variant_type) or \
            (self.position == other.position and self.variant_type == other.variant_type and self.wild_value < other.wild_value) or \
            (self.position == other.position and self.variant_type == other.variant_type and self.wild_value == other.wild_value and self.variant_value < other.variant_value)


    def __str__(self):
        return ''.join([self.wild_value, str(self.position + 1), self.variant_value])


    def sanity_check_against_seq(self, seq, translate_seq=False):
        if translate_seq:
            seq = pyfastaq.sequences.Fasta('x', seq).translate().seq

        return len(seq) >= self.position + 1 and seq[self.position].upper() in [self.wild_value, self.variant_value]


    def has_variant(self, seq):
        if self.variant_type == 'p':
            test_seq = seq.translate()
        else:
            test_seq = seq

        assert self.position < len(test_seq)
        return test_seq[self.position] == self.variant_value


    def nucleotide_range(self):
        '''Returns the nucleotide (start, end) positions inclusive of this variant.
           start==end if it's an amino acid variant, otherwise start+2==end'''
        if self.variant_type == 'p':
            return 3 * self.position, 3 * self.position + 2
        else:
            return self.position, self.position




import argparse
import subprocess
import shutil
import os
import sys
import ariba


def boxymcboxface(message):
    print('-' * 79)
    print('|', '=' * 77, '|', sep='')
    print('|', '{: ^75}'.format(message), '|')
    print('|', '=' * 77, '|', sep='')
    print('-' * 79)


def run(options):
    ariba_exe = os.path.abspath(sys.argv[0])

    print('Running ARIBA on test data...')

    boxymcboxface('Preparing input data')

    try:
        os.mkdir(options.outdir)
        os.chdir(options.outdir)
    except:
        print('Error making output directory "', options.outdir, '". Cannot continue.', sep='', file=sys.stderr)
        sys.exit(1)

    print('Made output directory ', options.outdir, '. Copying test data files into it:', sep='')

    modules_dir = os.path.dirname(os.path.abspath(ariba.__file__))
    test_data_dir = os.path.join(modules_dir, 'test_run_data')

    for filename in ['ref_seqs.fa', 'metadata.tsv', 'reads_1.fq', 'reads_2.fq']:
        shutil.copy(os.path.join(test_data_dir, filename), filename)
        print('    copied', filename)


    boxymcboxface('Try running ariba prepareref')

    prepareref_command = ' '.join([
        ariba_exe,
        'prepareref',
        '--verbose',
        '-f ref_seqs.fa',
        '-m metadata.tsv',
        '--threads', str(options.threads),
        'PREPAREREF',
    ])

    print('\nRunning ariba prepareref with:', prepareref_command, '', sep='\n')
    return_code = subprocess.call(prepareref_command, shell=True)

    if return_code != 0:
        print('\nSomething went wrong. See above for error message(s). Return code was', return_code)
        sys.exit(1)

    print()
    print('ariba prepareref finished OK')


    ariba_command = ' '.join([
        ariba_exe,
        'run',
        '--verbose',
        '--threads', str(options.threads),
        'PREPAREREF',
        'reads_1.fq',
        'reads_2.fq',
        'OUT'
    ])

    boxymcboxface('Try running ariba run')
    print('\nRunning ARIBA with:', ariba_command, '', sep='\n')

    return_code = subprocess.call(ariba_command, shell=True)

    if return_code != 0:
        print('\nSomething went wrong. See above for error message(s). Return code was', return_code)
        sys.exit(1)

    print()
    print('ariba run finished OK')
    print('Finished run on test data OK')



import os
import sys
import re
import copy
import pyfastaq
from ariba import sequence_metadata, cdhit


class Error (Exception): pass

rename_sub_regex = re.compile(r'''[|()\[\];"':!@,-]''')


class ReferenceData:
    def __init__(self,
        fasta_files,
        metadata_tsv_files,
        min_gene_length=6,
        max_gene_length=10000,
        genetic_code=11,
    ):
        self.seq_filenames = {}
        self.seq_dicts = {}
        self.min_gene_length = min_gene_length
        self.max_gene_length = max_gene_length

        self.sequences, self.metadata = ReferenceData._load_input_files_and_check_seq_names(fasta_files, metadata_tsv_files)
        if len(self.sequences) == 0:
            raise Error('Error. No sequences found in input file(s):' + '\n'.join(fasta_files) + '\nCannot continue')

        self.genetic_code = genetic_code
        pyfastaq.sequences.genetic_code = self.genetic_code
        self.rename_dict = None


    @classmethod
    def _load_metadata_tsv(cls, filename, metadata_dict):
        if filename is None:
            return {}

        f = pyfastaq.utils.open_file_read(filename)

        for line in f:
            try:
                metadata = sequence_metadata.SequenceMetadata(line)
            except:
                print('Problem with this line of metadata, which will be ignored:', line.rstrip(), file=sys.stderr)
                continue

            if metadata.name not in metadata_dict:
                metadata_dict[metadata.name] = {
                    'seq_type': metadata.seq_type,
                    'variant_only': metadata.variant_only,
                    'n': {},
                    'p': {},
                    '.': set()
                }
            elif metadata.seq_type != metadata_dict[metadata.name]['seq_type'] or metadata.variant_only != metadata_dict[metadata.name]['variant_only']:
                raise Error('Inconsistent metadata for sequence ' + metadata.name + '. Cannot continue')

            if metadata.variant is None:
                metadata_dict[metadata.name]['.'].add(metadata)
            else:
                if metadata.variant.position not in metadata_dict[metadata.name][metadata.seq_type]:
                    metadata_dict[metadata.name][metadata.seq_type][metadata.variant.position] = set()

                metadata_dict[metadata.name][metadata.seq_type][metadata.variant.position].add(metadata)

        pyfastaq.utils.close(f)
        return metadata_dict


    @classmethod
    def _load_all_metadata_tsvs(cls, filenames):
        metadata_dict = {}
        assert len(filenames) > 0
        for filename in filenames:
            ReferenceData._load_metadata_tsv(filename, metadata_dict)
        return metadata_dict


    @classmethod
    def _load_fasta_file(cls, filename, seq_dict):
        if filename is not None:
            seq_reader = pyfastaq.sequences.file_reader(filename)
            for seq in seq_reader:
                seq.id = seq.id.split()[0]

                if seq.id in seq_dict:
                    raise Error('Duplicate name "' + seq.id + '" found in file ' + filename + '. Cannot continue)')
                seq_dict[seq.id] = copy.copy(seq)


    @classmethod
    def _load_all_fasta_files(cls, filenames):
        seq_dict = {}
        assert len(filenames) > 0
        for filename in filenames:
            ReferenceData._load_fasta_file(filename, seq_dict)
        return seq_dict



    @staticmethod
    def _load_input_files_and_check_seq_names(fasta_files, metadata_files):
        metadata = ReferenceData._load_all_metadata_tsvs(metadata_files)
        all_seqs = ReferenceData._load_all_fasta_files(fasta_files)

        for seq_name in all_seqs:
            if seq_name not in metadata:
                raise Error('Sequence "' + seq_name + '" found in input fasta file but not in metadata file. Cannot continue')

        to_remove = set()

        for seq_name in metadata:
            if seq_name not in all_seqs:
                to_remove.add(seq_name)
                print('WARNING: sequence "', seq_name, '" found in metadata, but not in fasta file. Ignoring it.', sep='', file=sys.stderr)

        for key in to_remove:
            del metadata[key]

        return all_seqs, metadata


    @classmethod
    def _write_metadata_tsv(cls, metadata, filename):
        f = pyfastaq.utils.open_file_write(filename)

        for gene_name, data_dict in sorted(metadata.items()):
            for meta in sorted([str(x) for x in data_dict['.']]):
                print(meta, file=f)

            variants = []

            for variant_type in ['n', 'p']:
                for position in data_dict[variant_type]:
                    for meta in data_dict[variant_type][position]:
                        variants.append(meta)

            variants.sort()
            for v in variants:
                print(v, file=f)

        pyfastaq.utils.close(f)


    @classmethod
    def _write_sequences_to_files(cls, sequences, metadata, outprefix):
        filenames = {
            ('n', False): outprefix + '.noncoding.fa',
            ('n', True): outprefix + '.noncoding.varonly.fa',
            ('p', False): outprefix + '.gene.fa',
            ('p', True): outprefix + '.gene.varonly.fa',
        }

        filename2filehandle = {}
        for key in filenames:
            filename2filehandle[filenames[key]] = pyfastaq.utils.open_file_write(filenames[key])

        all_fh = pyfastaq.utils.open_file_write(outprefix + '.all.fa')

        for sequence in sorted(sequences):
            key = metadata[sequence]['seq_type'], metadata[sequence]['variant_only']
            filehandle = filename2filehandle[filenames[key]]
            print(sequences[sequence], file=filehandle)
            print(sequences[sequence], file=all_fh)

        for filehandle in filename2filehandle.values():
            pyfastaq.utils.close(filehandle)

        pyfastaq.utils.close(all_fh)


    @classmethod
    def _filter_bad_variant_data(cls, sequences, all_metadata, out_prefix, removed_sequences):
        genes_to_remove = set()
        variants_only_genes_found_variant = set()
        log_file = out_prefix + '.check_metadata.log'
        tsv_file = out_prefix + '.check_metadata.tsv'
        log_fh = pyfastaq.utils.open_file_write(log_file)

        for sequence_name, metadata_dict in sorted(all_metadata.items()):
            if sequence_name in removed_sequences:
                print(sequence_name, 'was removed because does not look like a gene, so removing its metadata', file=log_fh)
                del all_metadata[sequence_name]
                continue

            assert sequence_name in sequences

            # if this is non_coding, we shouldn't have any amino acid variants
            if metadata_dict['seq_type'] != 'p':
                for position in metadata_dict['p']:
                    for metadata in metadata_dict['p'][position]:
                        print(sequence_name, 'variant is an amino acid change, but sequence is non-coding. Removing. Line of file was:', metadata, file=log_fh)

                metadata_dict['p'] = {}

            # take out variant metadata that doesn't make sense (eg bases not matching ref sequence)
            for variant_type in ['n', 'p']:
                positions_to_remove = []
                for position in metadata_dict[variant_type]:
                    meta_to_remove = []
                    for metadata in metadata_dict[variant_type][position]:
                        to_translate = variant_type == 'p'

                        if not metadata.variant.sanity_check_against_seq(sequences[sequence_name], translate_seq=to_translate):
                            print(sequence_name, 'variant does not match reference. Removing. Line of file was:', metadata, file=log_fh)
                            meta_to_remove.append(metadata)
                            continue

                        if metadata_dict['variant_only']:
                            variants_only_genes_found_variant.add(sequence_name)

                    for metadata in meta_to_remove:
                        metadata_dict[variant_type][position].remove(metadata)
                    if len(metadata_dict[variant_type][position]) == 0:
                        positions_to_remove.append(position)

                for position in positions_to_remove:
                    del metadata_dict[variant_type][position]

            if metadata_dict['variant_only'] and len(metadata_dict['n']) == len(metadata_dict['p']) == len(metadata_dict['.']) == 0:
                print(sequence_name, 'No remaining data after checks. Removing this sequence because it is variants only', file=log_fh)
                genes_to_remove.add(sequence_name)

        for sequence_name in genes_to_remove:
            del all_metadata[sequence_name]
            del sequences[sequence_name]

        pyfastaq.utils.close(log_fh)
        ReferenceData._write_metadata_tsv(all_metadata, tsv_file)


    @classmethod
    def _try_to_get_gene_seq(cls, seq, min_length, max_length):
        seq.seq = seq.seq.upper()
        if len(seq) < min_length:
            return None, 'Remove: too short. Length: ' + str(len(seq))
        elif len(seq) > max_length:
            return None, 'Remove: too long. Length: ' + str(len(seq))
        else:
            got = seq.make_into_gene()
            if got is None:
                return None, 'Does not look like a gene (tried both strands and all reading frames) ' + seq.seq
            else:
                return got[0], 'Made ' + seq.id + ' into gene. strand=' + got[1] + ', frame=' + str(got[2])


    @classmethod
    def _remove_bad_genes(cls, sequences, metadata, log_file, min_gene_length, max_gene_length):
        to_remove = set()

        if len(sequences) == 0:
            return to_remove

        log_fh = pyfastaq.utils.open_file_write(log_file)

        for name in sorted(sequences):
            if metadata[name]['seq_type'] != 'p':
                continue

            new_seq, message = ReferenceData._try_to_get_gene_seq(sequences[name], min_gene_length, max_gene_length)
            if new_seq is None:
                to_remove.add(name)
            else:
                sequences[name] = new_seq

            if message is not None:
                print(name, message, file=log_fh)

        pyfastaq.utils.close(log_fh)

        for name in to_remove:
            sequences.pop(name)

        return to_remove


    def sanity_check(self, outprefix):
        removed_seqs = self._remove_bad_genes(self.sequences, self.metadata, outprefix + '.check_genes.log', self.min_gene_length, self.max_gene_length)
        ReferenceData._filter_bad_variant_data(self.sequences, self.metadata, outprefix, removed_seqs)


    @classmethod
    def _new_seq_name(cls, name):
        assert len(name.split()) == 1 and name.strip() == name
        return re.sub(rename_sub_regex, '_', name)


    @classmethod
    def _seq_names_to_rename_dict(cls, names):
        used_names = set()
        old_name_to_new = {}

        for old_name in sorted(names):
            assert len(old_name.split()) == 1 and old_name.strip() == old_name
            new_name = ReferenceData._new_seq_name(old_name)
            if new_name in used_names:
                i = 1
                new_name_prefix = new_name
                while new_name in used_names:
                    new_name = new_name_prefix + '_' + str(i)
                    i += 1

            assert new_name not in used_names
            if new_name != old_name:
                old_name_to_new[old_name] = new_name

            used_names.add(new_name)

        return old_name_to_new


    @classmethod
    def _rename_names_in_seq_dict(cls, seq_dict, rename_dict):
        new_dict = {}
        for name, seq in seq_dict.items():
            if name in rename_dict:
                seq.id = rename_dict[name]

            new_dict[seq.id] = seq
        return new_dict

    @classmethod
    def _rename_metadata_set(cls, metadata_set, new_name):
        new_set = set()
        for meta in metadata_set:
            new_meta = copy.copy(meta)
            new_meta.name = new_name
            new_set.add(new_meta)
        return new_set


    @classmethod
    def _rename_names_in_metadata(cls, meta_dict, rename_dict):
        new_dict = {}

        while len(meta_dict):
            old_name, gene_dict = meta_dict.popitem()
            if old_name in rename_dict:
                new_name = rename_dict[old_name]
                for seq_type in ['n', 'p']:
                    for position, metaset in gene_dict[seq_type].items():
                        gene_dict[seq_type][position] = ReferenceData._rename_metadata_set(metaset, new_name)

                gene_dict['.'] = ReferenceData._rename_metadata_set(gene_dict['.'], new_name)
            else:
                new_name = old_name

            new_dict[new_name] = gene_dict

        return new_dict


    def rename_sequences(self, outfile):
        self.rename_dict = ReferenceData._seq_names_to_rename_dict(self.sequences.keys())
        if len(self.rename_dict):
            print('Had to rename some sequences. See', outfile, 'for old -> new names', file=sys.stderr)
            with open(outfile, 'w') as f:
                for old_name, new_name in sorted(self.rename_dict.items()):
                    print(old_name, new_name, sep='\t', file=f)

            self.sequences = ReferenceData._rename_names_in_seq_dict(self.sequences, self.rename_dict)
            self.metadata = ReferenceData._rename_names_in_metadata(self.metadata, self.rename_dict)


    def sequence_type(self, sequence_name):
        assert sequence_name in self.metadata
        return self.metadata[sequence_name]['seq_type'], self.metadata[sequence_name]['variant_only']


    def sequence(self, sequence_name):
        return self.sequences.get(sequence_name, None)


    def all_non_wild_type_variants(self, ref_name):
        ref_seq = self.sequence(ref_name)
        variants = {'n': {}, 'p': {}}

        if ref_seq is None or ref_name not in self.metadata:
            return variants

        for variant_type in ['n', 'p']:
            for position, metadata_set in self.metadata[ref_name][variant_type].items():
                for metadata in metadata_set:
                    if position not in variants[variant_type]:
                        variants[variant_type][position] = set()

                    variants[variant_type][position].add(metadata)

        return variants


    @staticmethod
    def write_cluster_allocation_file(clusters, outfile):
        f_out = pyfastaq.utils.open_file_write(outfile)

        for cluster, name_set in sorted(clusters.items()):
            seq_names = sorted(list(name_set))
            print(cluster, *seq_names, sep='\t', file=f_out)

        pyfastaq.utils.close(f_out)


    def cluster_with_cdhit(self, outprefix, seq_identity_threshold=0.9, threads=1, length_diff_cutoff=0.9, nocluster=False, verbose=False, clusters_file=None):
        clusters = {}
        ReferenceData._write_sequences_to_files(self.sequences, self.metadata, outprefix)
        ref_types = ('noncoding', 'noncoding.varonly', 'gene', 'gene.varonly')

        for ref_type in ref_types:
            ref_file = outprefix + '.' + ref_type + '.fa'
            if os.path.getsize(ref_file) == 0:
                continue

            if len(clusters) == 0:
                min_cluster_number = 0
            else:
                min_cluster_number = 1 + max([int(x) for x in clusters.keys()])

            cdhit_runner = cdhit.Runner(
              ref_file,
              seq_identity_threshold=seq_identity_threshold,
              threads=threads,
              length_diff_cutoff=length_diff_cutoff,
              verbose=verbose,
              min_cluster_number = min_cluster_number,
            )

            if clusters_file is not None:
                new_clusters = cdhit_runner.run_get_clusters_from_file(clusters_file, self.sequences, rename_dict=self.rename_dict)
            elif nocluster:
                new_clusters = cdhit_runner.fake_run()
            else:
                new_clusters = cdhit_runner.run()

            clusters.update(new_clusters)

        self.write_cluster_allocation_file(clusters, outprefix + '.clusters.tsv')
        return clusters


    def write_seqs_to_fasta(self, outfile, names):
        f_out = pyfastaq.utils.open_file_write(outfile)

        for name in sorted(names):
            print(self.sequence(name), file=f_out)

        pyfastaq.utils.close(f_out)



import pyfastaq
from ariba import link

class Error (Exception): pass

class Graph:
    def __init__(self, ref_lengths):
        self.links = {}
        self.partial_links = {}
        self.ref_lengths = ref_lengths


    def update_from_sam(self, sam, sam_reader):
        '''Updates graph info from a pysam.AlignedSegment object'''
        if sam.is_unmapped \
          or sam.mate_is_unmapped \
          or (sam.reference_id == sam.next_reference_id):
            return

        new_link = link.Link(sam, sam_reader, self.ref_lengths)
        read_name = sam.query_name

        if read_name in self.partial_links:
            new_link.merge(self.partial_links[read_name])
            del self.partial_links[read_name]
            key = tuple(sorted((new_link.refnames[0], new_link.refnames[1])))
            if key not in self.links:
                self.links[key] = []
            new_link.sort()
            self.links[key].append(new_link)
        else:
            self.partial_links[read_name] = new_link


    def _make_graph(self, max_insert):
        '''helper function to construct graph from current state of object'''
        if len(self.partial_links) != 0:
            raise Error('Error in _make_graph(). Cannot continue because there are partial links')
        
        self.contig_links = {}
        for key in self.links:
            for l in self.links[key]:
                insert_size = l.insert_size()
                if insert_size <= max_insert:
                    if key not in self.contig_links:
                        self.contig_links[key] = {}
                    dirs = ''.join(l.dirs)
                    self.contig_links[key][dirs] = self.contig_links[key].get(dirs, 0) + 1


    def _remove_low_cov_links(self, min_coverage):
        keys_to_delete = set()

        for key in self.contig_links:
            d = self.contig_links[key]
            self.contig_links[key] = {x:d[x] for x in d if d[x] >= min_coverage}
            if len(self.contig_links[key]) == 0:
                keys_to_delete.add(key)

        for key in keys_to_delete:
            del self.contig_links[key]


    def _contig_graph_is_consistent(self):
        for key in self.contig_links:
            if len(self.contig_links[key]) > 1:
                return False

        all_names_in_keys = set()
        for key in self.contig_links:
            all_names_in_keys.add(key[0])
            all_names_in_keys.add(key[1])

        name_to_links = {name: set() for name in all_names_in_keys}

        for key in self.contig_links:
            assert len(self.contig_links[key]) == 1
            directions = list(self.contig_links[key])[0]
            name1, name2 = key
            dir1, dir2 = directions[0], directions[1]
            for n, d in [(name1, dir1), (name2, dir2)]:
                if d in name_to_links[n]:
                    return False
                else:
                    name_to_links[n].add(d)
        return True


    def is_consistent(self, min_coverage, max_insert):
        self._make_graph(max_insert)
        self._remove_low_cov_links(min_coverage)
        return self._contig_graph_is_consistent()


    def write_all_links_to_file(self, filename):
        f = pyfastaq.utils.open_file_write(filename)
        for key in sorted(self.links):
            to_write = list(self.links[key])
            to_write.sort()
            for l in to_write:
                print(l, file=f)
        pyfastaq.utils.close(f)




import os
import sys
import shutil
import pyfastaq
import pymummer
import fermilite_ariba
from ariba import common, faidx, mapping, bam_parse, external_progs, mash

class Error (Exception): pass

class Assembly:
    def __init__(self,
      reads1,
      reads2,
      ref_fasta,
      ref_fastas,
      working_dir,
      final_assembly_fa,
      final_assembly_bam,
      log_fh,
      scaff_name_prefix='scaffold',
      kmer=0,
      assembler='fermilite',
      max_insert=1000,
      min_scaff_depth=10,
      min_scaff_length=50,
      nucmer_min_id=90,
      nucmer_min_len=20,
      nucmer_breaklen=200,
      spades_other_options=None,
      sspace_k=20,
      sspace_sd=0.4,
      reads_insert=500,
      extern_progs=None,
      clean=True,
    ):
        self.reads1 = os.path.abspath(reads1)
        self.reads2 = os.path.abspath(reads2)
        self.ref_fasta = os.path.abspath(ref_fasta)
        self.ref_fastas = os.path.abspath(ref_fastas)
        self.working_dir = os.path.abspath(working_dir)
        self.final_assembly_fa = os.path.abspath(final_assembly_fa)
        self.final_assembly_bam = os.path.abspath(final_assembly_bam)
        self.log_fh = log_fh
        self.scaff_name_prefix = scaff_name_prefix

        self.ref_seq_name = None
        self.assembly_kmer = self._get_assembly_kmer(kmer, reads1, reads2)
        self.assembler = assembler
        self.max_insert = max_insert
        self.min_scaff_depth = min_scaff_depth
        self.min_scaff_length = min_scaff_length
        self.nucmer_min_id = nucmer_min_id
        self.nucmer_min_len = nucmer_min_len
        self.nucmer_breaklen = nucmer_breaklen
        self.spades_other_options = spades_other_options
        self.sspace_k = sspace_k
        self.sspace_sd = sspace_sd
        self.reads_insert = reads_insert
        self.clean = clean

        if extern_progs is None:
            self.extern_progs = external_progs.ExternalProgs()
        else:
            self.extern_progs = extern_progs

        try:
            os.mkdir(self.working_dir)
        except:
            raise Error('Error mkdir ' + self.working_dir)

        self.assembler_dir = os.path.join(self.working_dir, 'Assemble')
        self.assembly_contigs = os.path.join(self.working_dir, 'contigs.fa')
        self.scaffold_dir = os.path.join(self.working_dir, 'Scaffold')
        self.scaffolder_scaffolds = os.path.join(self.working_dir, 'scaffolds.fa')
        self.gapfill_dir = os.path.join(self.working_dir, 'Gapfill')
        self.gapfilled_scaffolds = os.path.join(self.working_dir, 'scaffolds.gapfilled.fa')
        self.gapfilled_length_filtered = os.path.join(self.working_dir, 'scaffolds.gapfilled.length_filtered.fa')
        self.mash_dist_file = os.path.join(self.working_dir, 'mash_dist_to_ref_seqs')


    @staticmethod
    def _get_assembly_kmer(k, reads1, reads2):
        '''If the kmer not given, uses 2/3 of the mean read length (using first 1000 forward and first 1000 reverse reads)'''
        if k == 0:
            read_length1 = pyfastaq.tasks.mean_length(reads1, limit=1000)
            read_length2 = pyfastaq.tasks.mean_length(reads2, limit=1000)
            assembly_kmer = round( (read_length1 + read_length2) / 3)
            if assembly_kmer % 2 == 0:
                assembly_kmer += 1
        else:
            assembly_kmer = k

        return assembly_kmer


    @staticmethod
    def _check_spades_log_file(logfile):
        '''SPAdes can fail with a strange error. Stop everything if this happens'''
        f = pyfastaq.utils.open_file_read(logfile)

        for line in f:
            if line.startswith('== Error ==  system call for:') and line.rstrip().endswith('finished abnormally, err code: -7'):
                pyfastaq.utils.close(f)
                print('Error running SPAdes. Cannot continue. This is the error from the log file', logfile, '...', file=sys.stderr)
                print(line, file=sys.stderr)
                raise Error('Fatal error ("err code: -7") running spades. Cannot continue')

        pyfastaq.utils.close(f)
        return True


    @staticmethod
    def _run_fermilite(reads_in, fasta_out, log_out):
        return fermilite_ariba.fermilite_ariba(reads_in, fasta_out, log_out)


    def _assemble_with_fermilite(self):
        cwd = os.getcwd()
        try:
            os.chdir(self.working_dir)
        except:
            raise Error('Error chdir ' + self.working_dir)

        interleaved_reads = 'reads.fq'
        pyfastaq.tasks.interleave(self.reads1, self.reads2, interleaved_reads)
        fermilite_log = self.assembly_contigs + '.log'
        got_from_fermilite = self._run_fermilite(interleaved_reads, self.assembly_contigs, fermilite_log)
        os.unlink(interleaved_reads)
        if os.path.exists(fermilite_log):
            with open(fermilite_log) as f:
                for line in f:
                    print(line, end='', file=self.log_fh)

            os.unlink(fermilite_log)

        self.assembled_ok = (got_from_fermilite == 0)
        os.chdir(cwd)


    def _assemble_with_spades(self, unittest=False):
        cmd = ' '.join([
            self.extern_progs.exe('spades'),
            '-1', self.reads1,
            '-2', self.reads2,
            '-o', self.assembler_dir,
            '-k', str(self.assembly_kmer),
            '--threads 1', # otherwise defaults to 16!
        ])
        if self.spades_other_options is not None:
            cmd += ' ' + self.spades_other_options

        cwd = os.getcwd()
        try:
            os.chdir(self.working_dir)
        except:
            raise Error('Error chdir ' + self.working_dir)
        spades_contigs = os.path.join(os.path.split(self.assembler_dir)[1], 'scaffolds.fasta')

        if unittest:
            os.mkdir(self.assembler_dir)
            open(spades_contigs, 'w').close()
            self.assembled_ok = True
        else:
            self.assembled_ok, err = common.syscall(cmd, verbose=True, allow_fail=True, verbose_filehandle=self.log_fh, print_errors=False)
        if self.assembled_ok:
            os.rename(spades_contigs, os.path.basename(self.assembly_contigs))
        else:
            print('Assembly finished with errors. These are the errors:', file=self.log_fh)
            print(err, file=self.log_fh)
            print('\nEnd of spades errors\n', file=self.log_fh)

        spades_log = os.path.join(self.assembler_dir, 'spades.log')
        if os.path.exists(spades_log):
            self._check_spades_log_file(spades_log)

            with open(spades_log) as f:
                print('\n______________ SPAdes log ___________________\n', file=self.log_fh)
                for line in f:
                    print(line.rstrip(), file=self.log_fh)
                print('\n______________ End of SPAdes log _________________\n', file=self.log_fh)


        spades_warnings = os.path.join(self.assembler_dir, 'warnings.log')
        if os.path.exists(spades_warnings):
            with open(spades_warnings) as f:
                print('\n______________ SPAdes warnings ___________________\n', file=self.log_fh)
                for line in f:
                    print(line.rstrip(), file=self.log_fh)
                print('\n______________ End of SPAdes warnings _________________\n', file=self.log_fh)

        os.chdir(cwd)

        if self.clean:
            print('Deleting assembly directory', self.assembler_dir, file=self.log_fh)
            shutil.rmtree(self.assembler_dir)


    def _scaffold_with_sspace(self):
        if not os.path.exists(self.assembly_contigs):
            raise Error('Cannot scaffold because contigs file not found: ' + self.assembly_contigs)

        try:
            os.mkdir(self.scaffold_dir)
        except:
            raise Error('Error mkdir '+  self.scaffold_dir)

        cwd = os.getcwd()

        #if self.extern_progs.exe('sspace') is None:
        if True:  # no longer use sspace, but leave the option here just in case
            os.chdir(self.working_dir)
            os.symlink(self.assembly_contigs, os.path.basename(self.scaffolder_scaffolds))
            os.chdir(cwd)
            return

        os.chdir(self.scaffold_dir)
        lib_file = 'lib'
        with open(lib_file, 'w') as f:
            print('LIB', self.reads1, self.reads2, int(self.reads_insert), self.sspace_sd, 'FR', file=f)

        cmd = ' '.join([
            'perl', self.extern_progs.exe('sspace'),
            '-k', str(self.sspace_k),
            '-l', lib_file,
            '-s', self.assembly_contigs
        ])

        common.syscall(cmd, verbose=True, verbose_filehandle=self.log_fh)
        sspace_scaffolds = os.path.abspath('standard_output.final.scaffolds.fasta')
        sspace_log = os.path.abspath('standard_output.logfile.txt')
        with open(sspace_log) as f:
            print('\n_______________ SSPACE log __________________\n', file=self.log_fh)
            for line in f:
                print(line.rstrip(), file=self.log_fh)
            print('_______________ End of SSPACE log __________________\n', file=self.log_fh)

        os.rename(sspace_scaffolds, self.scaffolder_scaffolds)
        os.chdir(cwd)

        if self.clean:
            print('Deleting scaffolding directory', self.scaffold_dir, file=self.log_fh)
            shutil.rmtree(self.scaffold_dir)


    @staticmethod
    def _has_gaps_to_fill(filename):
        seq_reader = pyfastaq.sequences.file_reader(filename)
        for seq in seq_reader:
            if 'n' in seq.seq or 'N' in seq.seq:
                return True
        return False


    @staticmethod
    def _rename_scaffolds(infile, outfile, prefix):
        freader = pyfastaq.sequences.file_reader(infile)
        f_out = pyfastaq.utils.open_file_write(outfile)
        i = 1
        for scaff in freader:
            scaff.id = prefix + '.scaffold.' + str(i)
            i += 1
            print(scaff, file=f_out)
        pyfastaq.utils.close(f_out)


    def _gap_fill_with_gapfiller(self):
        if not os.path.exists(self.scaffolder_scaffolds):
            raise Error('Cannot gap fill because scaffolds file not found: ' + self.scaffolder_scaffolds)

        cwd = os.getcwd()

        #if self.extern_progs.exe('gapfiller') is None or not self._has_gaps_to_fill(self.scaffolder_scaffolds):
        if True: # no longer use gapfiller, but leave the option in, just in case
            self._rename_scaffolds(self.scaffolder_scaffolds, self.gapfilled_scaffolds, self.scaff_name_prefix)
            return

        try:
            os.mkdir(self.gapfill_dir)
        except:
            raise Error('Error mkdir '+  self.gapfill_dir)

        os.chdir(self.gapfill_dir)
        lib_file = 'lib'
        with open(lib_file, 'w') as f:
            print('LIB', 'bwa', self.reads1, self.reads2, self.reads_insert, self.sspace_sd, 'FR', file=f)

        cmd = ' '.join([
            'perl', self.extern_progs.exe('gapfiller'),
            '-l', lib_file,
            '-s', self.scaffolder_scaffolds
        ])

        gapfilled_scaffolds = os.path.join(self.gapfill_dir, 'standard_output', 'standard_output.gapfilled.final.fa')
        common.syscall(cmd, verbose=True, verbose_filehandle=self.log_fh)
        self._rename_scaffolds(gapfilled_scaffolds, self.gapfilled_scaffolds, self.scaff_name_prefix)
        os.chdir(cwd)
        if self.clean:
            print('Deleting GapFiller directory', self.gapfill_dir, file=self.log_fh)
            shutil.rmtree(self.gapfill_dir)


    @staticmethod
    def _fix_contig_orientation(contigs_fa, ref_fa, outfile, min_id=90, min_length=20, breaklen=200):
        '''Changes orientation of each contig to match the reference, when possible.
           Returns a set of names of contigs that had hits in both orientations to the reference'''
        if not os.path.exists(contigs_fa):
            raise Error('Cannot fix orientation of assembly contigs because file not found: ' + contigs_fa)

        tmp_coords = os.path.join(outfile + '.tmp.rename.coords')
        pymummer.nucmer.Runner(
            ref_fa,
            contigs_fa,
            tmp_coords,
            min_id=min_id,
            min_length=min_length,
            breaklen=breaklen,
            maxmatch=True,
        ).run()

        to_revcomp = set()
        not_revcomp = set()
        file_reader = pymummer.coords_file.reader(tmp_coords)
        for hit in file_reader:
            if hit.on_same_strand():
                not_revcomp.add(hit.qry_name)
            else:
                to_revcomp.add(hit.qry_name)

        os.unlink(tmp_coords)
        in_both = to_revcomp.intersection(not_revcomp)

        f = pyfastaq.utils.open_file_write(outfile)
        seq_reader = pyfastaq.sequences.file_reader(contigs_fa)
        for seq in seq_reader:
            if seq.id in to_revcomp and seq.id not in in_both:
                seq.revcomp()
            print(seq, file=f)
        pyfastaq.utils.close(f)

        return in_both


    @staticmethod
    def _parse_bam(sequences, bam, min_scaff_depth, max_insert):
        if not os.path.exists(bam):
            raise Error('File not found: ' + bam)

        bam_parser = bam_parse.Parser(bam, sequences)
        bam_parser.parse()
        bam_parser.write_files(bam)
        return bam_parser.scaff_graph_is_consistent(min_scaff_depth, max_insert)


    def run(self):
        self._assemble_with_fermilite()
        self.sequences = {}

        # double-check we got some contigs
        number_of_contigs = pyfastaq.tasks.count_sequences(self.assembly_contigs) if os.path.exists(self.assembly_contigs) else 0
        if number_of_contigs == 0:
            self.assembled_ok = False
            # This is to make this object picklable, to keep multithreading happy
            self.log_fh = None
            return
        else:
            self.assembled_ok = True

        if self.assembled_ok:
            self._scaffold_with_sspace()
            self._gap_fill_with_gapfiller()

            pyfastaq.tasks.filter(self.gapfilled_scaffolds, self.gapfilled_length_filtered, minlength=self.min_scaff_length)
            if pyfastaq.tasks.count_sequences(self.gapfilled_length_filtered) == 0:
                self.assembled_ok = False
                # This is to make this object picklable, to keep multithreading happy
                self.log_fh = None
                return

            masher = mash.Masher(self.ref_fastas, self.gapfilled_length_filtered, self.log_fh, self.extern_progs)
            self.ref_seq_name = masher.run(self.mash_dist_file)
            if self.ref_seq_name is None:
                print('Could not determine closest reference sequence', file=self.log_fh)
                self.log_fh = None
                return

            faidx.write_fa_subset({self.ref_seq_name}, self.ref_fastas, self.ref_fasta, samtools_exe=self.extern_progs.exe('samtools'), verbose=True, verbose_filehandle=self.log_fh)
            print('Closest reference sequence according to mash: ', self.ref_seq_name, file=self.log_fh)

            contigs_both_strands = self._fix_contig_orientation(self.gapfilled_length_filtered, self.ref_fasta, self.final_assembly_fa, min_id=self.nucmer_min_id, min_length=self.nucmer_min_len, breaklen=self.nucmer_breaklen)
            self.has_contigs_on_both_strands = len(contigs_both_strands) > 0
            pyfastaq.tasks.file_to_dict(self.final_assembly_fa, self.sequences)

            mapping.run_bowtie2(
                self.reads1,
                self.reads2,
                self.final_assembly_fa,
                self.final_assembly_bam[:-4],
                threads=1,
                sort=True,
                samtools=self.extern_progs.exe('samtools'),
                bowtie2=self.extern_progs.exe('bowtie2'),
                verbose=True,
                verbose_filehandle=self.log_fh
            )

            self.scaff_graph_ok = self._parse_bam(self.sequences, self.final_assembly_bam, self.min_scaff_depth, self.max_insert)
            print('Scaffolding graph is OK:', self.scaff_graph_ok, file=self.log_fh)

            if self.clean:
                for suffix in ['soft_clipped', 'unmapped_mates', 'scaff']:
                    filename = self.final_assembly_bam + '.' + suffix
                    print('Deleting file', filename, file=self.log_fh)
                    os.unlink(filename)


        # This is to make this object picklable, to keep multithreading happy
        self.log_fh = None



import shutil
import tempfile
import os
import pyfastaq
from ariba import mapping, faidx

class Error (Exception): pass

class BestSeqChooser:
    def __init__(self,
        reads1,
        reads2,
        references_fa,
        log_fh,
        samtools_exe='samtools',
        bowtie2_exe='bowtie2',
        bowtie2_preset='very-sensitive-local',
        threads=1,
    ):
        self.reads1 = reads1
        self.reads2 = reads2
        self.references_fa = references_fa
        self.log_fh = log_fh
        self.samtools_exe = samtools_exe
        self.bowtie2_exe = bowtie2_exe
        self.bowtie2_preset = bowtie2_preset
        self.threads = threads


    def _total_alignment_score(self, seq_name):
        tmpdir = tempfile.mkdtemp(prefix='tmp.get_total_aln_score.', dir=os.getcwd())
        tmp_bam = os.path.join(tmpdir, 'tmp.get_total_alignment_score.bam')
        tmp_fa = os.path.join(tmpdir, 'tmp.get_total_alignment_score.ref.fa')

        faidx.write_fa_subset(
            [seq_name],
            self.references_fa,
            tmp_fa,
            samtools_exe=self.samtools_exe,
            verbose=True,
            verbose_filehandle=self.log_fh
        )

        mapping.run_bowtie2(
            self.reads1,
            self.reads2,
            tmp_fa,
            tmp_bam[:-4],
            threads=self.threads,
            samtools=self.samtools_exe,
            bowtie2=self.bowtie2_exe,
            bowtie2_preset=self.bowtie2_preset,
            verbose=True,
            verbose_filehandle=self.log_fh
        )

        score = mapping.get_total_alignment_score(tmp_bam)
        shutil.rmtree(tmpdir)
        return score


    def _get_best_seq_by_alignment_score(self):
        total_sequences = pyfastaq.tasks.count_sequences(self.references_fa)
        if total_sequences == 1:
            seqs = {}
            pyfastaq.tasks.file_to_dict(self.references_fa, seqs)
            assert len(seqs) == 1
            seq_name = list(seqs.values())[0].id
            print('No need to choose sequence for this cluster because only has one sequence:', seq_name, file=self.log_fh)
            return seq_name

        print('\nChoosing best sequence from cluster of', total_sequences, 'sequences...', file=self.log_fh)
        file_reader = pyfastaq.sequences.file_reader(self.references_fa)
        best_score = 0
        best_seq_name = None
        for seq in file_reader:
            score = self._total_alignment_score(seq.id)
            print('Total alignment score for sequence', seq.id, 'is', score, file=self.log_fh)
            if score > best_score:
                best_score = score
                best_seq_name = seq.id

        print('\nBest sequence is', best_seq_name, 'with total alignment score of', best_score, file=self.log_fh)
        print(file=self.log_fh)
        return best_seq_name


    def best_seq(self, outfile):
        '''Finds the closest matchng sequence, writes it to a FASTA file, and returns it as a pyfastaq.sequences.Fasta object'''
        seq_name = self._get_best_seq_by_alignment_score()
        if seq_name is None:
            return None
        faidx.write_fa_subset([seq_name], self.references_fa, outfile, samtools_exe=self.samtools_exe, verbose=True, verbose_filehandle=self.log_fh)
        seqs = {}
        pyfastaq.tasks.file_to_dict(outfile, seqs)
        assert len(seqs) == 1
        return list(seqs.values())[0]




import os
import sys
import pysam
import pyfastaq
from ariba import common

class Error (Exception): pass

bowtie2_index_extensions = [x + '.bt2' for x in ['1', '2', '3', '4', 'rev.1', 'rev.2']]

def bowtie2_index(ref_fa, outprefix, bowtie2='bowtie2', verbose=False, verbose_filehandle=sys.stdout):
    expected_files = [outprefix + '.' + x + '.bt2' for x in ['1', '2', '3', '4', 'rev.1', 'rev.2']]
    file_missing = False
    for filename in expected_files:
        if not os.path.exists(filename):
            file_missing = True
            break

    if not file_missing:
        return

    cmd = ' '.join([
        bowtie2 + '-build',
        '-q',
        ref_fa,
        outprefix
    ])

    common.syscall(cmd, verbose=verbose, verbose_filehandle=verbose_filehandle)


def run_bowtie2(
      reads_fwd,
      reads_rev,
      ref_fa,
      out_prefix,
      threads=1,
      max_insert=1000,
      sort=False,
      samtools='samtools',
      bowtie2='bowtie2',
      bowtie2_preset='very-sensitive-local',
      verbose=False,
      verbose_filehandle=sys.stdout,
      remove_both_unmapped=False,
      clean_index=True,
    ):

    ref_is_indexed = True
    for ext in bowtie2_index_extensions:
        if not os.path.exists(ref_fa + '.' + ext):
            ref_is_indexed = False
            break

    clean_files = []

    if ref_is_indexed:
        if verbose:
            print('Bowtie2 index files found (', ref_fa, '.*.bt2) so no need to index', sep='', file=verbose_filehandle)
        map_index = ref_fa
    else:
        map_index = out_prefix + '.map_index'
        bowtie2_index(ref_fa, map_index, bowtie2=bowtie2, verbose=verbose, verbose_filehandle=verbose_filehandle)

        if clean_index:
            clean_files = [map_index + '.' + x + '.bt2' for x in ['1', '2', '3', '4', 'rev.1', 'rev.2']]

    final_bam = out_prefix + '.bam'
    if sort:
        intermediate_bam = out_prefix + '.unsorted.bam'
    else:
        intermediate_bam = final_bam

    map_cmd = [
        bowtie2,
        '--threads', str(threads),
        '--reorder',
        '--' + bowtie2_preset,
        '-X', str(max_insert),
        '-x', map_index,
        '-1', reads_fwd,
        '-2', reads_rev,
    ]

    if remove_both_unmapped:
        map_cmd.append(r''' | awk ' !(and($2,4)) || !(and($2,8)) ' ''')


    map_cmd.extend([
        '|', samtools, 'view',
        '-bS -T', ref_fa,
        '- >', intermediate_bam
    ])

    map_cmd = ' '.join(map_cmd)

    common.syscall(map_cmd, verbose=verbose, verbose_filehandle=verbose_filehandle)

    if sort:
        threads = min(4, threads)
        thread_mem = int(500 / threads)
        sort_cmd = ' '.join([
            samtools,
            'sort',
            '-@' + str(threads),
            '-m' + str(thread_mem) + 'M',
            '-o', final_bam,
            '-O bam',
            '-T', out_prefix + '.tmp.samtool_sort',
            intermediate_bam,
        ])
        index_cmd = samtools + ' index ' + final_bam
        common.syscall(sort_cmd, verbose=verbose, verbose_filehandle=verbose_filehandle)
        common.syscall(index_cmd, verbose=verbose, verbose_filehandle=verbose_filehandle)
        clean_files.append(intermediate_bam)

    for fname in clean_files:
        os.unlink(fname)


def get_total_alignment_score(bam):
    '''Returns total of AS: tags in the input BAM'''
    sam_reader = pysam.Samfile(bam, "rb")
    total = 0
    for sam in sam_reader.fetch(until_eof=True):
        try:
            total += sam.opt('AS')
        except:
            pass
    return total


def sam_to_fastq(sam):
    '''Given a pysam alignment, returns the sequence a Fastq object.
       Reverse complements as required and add suffix /1 or /2 as appropriate from the flag'''
    name = sam.qname
    if sam.is_read1:
        name += '/1'
    elif sam.is_read2:
        name += '/2'
    else:
        raise Error('Read ' + name + ' must be first or second of pair according to flag. Cannot continue')

    seq = pyfastaq.sequences.Fastq(name, common.decode(sam.seq), common.decode(sam.qual))
    if sam.is_reverse:
        seq.revcomp()

    return seq


def sam_pair_to_insert(s1, s2):
    '''Returns insert size from pair of sam records, as long as their orientation is "innies".
       Otherwise returns None.'''
    if s1.is_unmapped or s2.is_unmapped or (s1.tid != s2.tid) or (s1.is_reverse == s2.is_reverse):
        return None

    # If here, reads are both mapped to the same ref, and in opposite orientations
    if s1.is_reverse:
        end = s1.reference_end - 1
        start = s2.reference_start
    else:
        end = s2.reference_end - 1
        start = s1.reference_start

    if start < end:
        return end - start + 1
    else:
        return None





import argparse
from ariba import aln_to_metadata


def run(options):
    aln_to_meta = aln_to_metadata.AlnToMetadata(
      options.aln_fasta,
      options.variants_tsv,
      options.coding_or_non == 'coding',
      options.variant_only,
      genetic_code=options.genetic_code
    )
    aln_to_meta.run(options.outprefix)

